"""setup.py"""
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="altimeter",
    version="1.0.27-rc2",
    packages=find_packages(exclude=["tests"]),
    author="Tableau",
    description="Graph AWS resources in Neptune",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tableau/altimeter",
    python_requires=">=3.7,<4",
    install_requires=[
        "aws-requests-auth==0.4.2",
        "rdflib==4.2.2",
        "structlog>=18.2.0,<20",
        "boto3>=1.9.130",
        "typing_extensions>=3.7.4.1,<3.8",
        "jinja2>=2.11.1,<3",
    ],
    scripts=[
        "bin/account_scan.py",
        "bin/altimeter",
        "bin/aws2json.py",
        "bin/graphpruner.py",
        "bin/json2rdf.py",
        "bin/rdf2blaze",
        "bin/rdf2n.py",
        "bin/runquery.py",
        "bin/scan_resource.py",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Tags Fields represent AWS Tags."""
from typing import Dict, Any, List

from altimeter.core.graph.field.base import Field
from altimeter.core.graph.field.exceptions import TagsFieldMissingTagsKeyException
from altimeter.core.graph.link.links import TagLink
from altimeter.core.graph.link.base import Link


class TagsField(Field):
    """A TagsField is a field used to parse data in the form

        {'Tags': [{'Key': 'tag_key', 'Value': 'tag_value'}, ...]}

    Examples:
        >>> input = {"Tags": [{"Key": "Name", "Value": "Jerry"}, \
                              {"Key": "DOB", "Value": "1942-08-01"}]}
        >>> field = TagsField()
        >>> links = field.parse(data=input, context={})
        >>> print(links[0].to_dict())
        {'pred': 'Name', 'obj': 'Jerry', 'type': 'tag'}
        >>> print(links[1].to_dict())
        {'pred': 'DOB', 'obj': '1942-08-01', 'type': 'tag'}

    Args:
        optional: Whether this key is optional. Defaults to False.
    """

    def __init__(self, optional: bool = True):
        self.optional = optional

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

       Args:
           data: dictionary of data to parse
           context: context dict containing data from higher level parsing code.

        Returns:
            List of TagLink objects, one for each tag.
        """
        fields: List[Link] = []
        tag_dicts = data.get("Tags")
        if tag_dicts:
            for tag_dict in tag_dicts:
                field = TagLink(pred=tag_dict["Key"], obj=tag_dict["Value"])
                fields.append(field)
            return fields
        if self.optional:
            return []
        raise TagsFieldMissingTagsKeyException(f"Expected key 'Tags' in {data}")



"""Field parsing related exceptions."""
from altimeter.core.exceptions import AltimeterException


class ScalarFieldSourceKeyNotFoundException(AltimeterException):
    """The source_key of a ScalarField is not found"""


class ScalarFieldValueNotAScalarException(AltimeterException):
    """The value of a ScalarField is not a string, bool, int or float."""


class ResourceLinkFieldSourceKeyNotFoundException(AltimeterException):
    """The source_key of a ResourceLinkField is not found"""


class ResourceLinkFieldValueNotAStringException(AltimeterException):
    """The value of a ResourceLinkField is not a string."""


class TagsFieldMissingTagsKeyException(AltimeterException):
    """A TagsField data is missing key 'Tags'"""


class ParentKeyMissingException(AltimeterException):
    """A required ParentKey is missing."""


class InvalidParentKeyException(AltimeterException):
    """A ParentKey is invalid."""


class DictFieldValueNotADictException(AltimeterException):
    """A DictField does not contain a dict."""


class DictFieldSourceKeyNotFoundException(AltimeterException):
    """The source_key of a DictField was not found"""


class DictFieldValueIsNullException(AltimeterException):
    """The value of a non-nullable DictField was null"""



"""Base class for DynamoDB resources."""

from altimeter.aws.resource.resource_spec import AWSResourceSpec


class DynamoDBResourceSpec(AWSResourceSpec):
    """Base class for DynamoDB resources."""

    service_name = "dynamodb"



"""Resource representing an AWS Account as viewed in Orgs. This tags
on things like the Org itself and the OU in which this account lives."""
from typing import Any, Dict, List, Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.organizations import OrganizationsResourceSpec
from altimeter.aws.resource.organizations.org import OrgResourceSpec
from altimeter.aws.resource.organizations.ou import OUResourceSpec
from altimeter.aws.resource.unscanned_account import UnscannedAccountResourceSpec
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class OrgsAccountResourceSpec(OrganizationsResourceSpec):
    """Resource representing an AWS Account as viewed in Orgs."""

    type_name = "account"
    schema = Schema(
        ScalarField("Name"),
        ScalarField("Status"),
        ScalarField("JoinedTimestamp"),
        ScalarField("Email"),
        ScalarField("Id", alti_key="account_id"),
        ResourceLinkField("OrganizationArn", OrgResourceSpec, value_is_id=True),
        ResourceLinkField("OUArn", OUResourceSpec, value_is_id=True),
    )
    allow_clobber = [UnscannedAccountResourceSpec]

    @classmethod
    def get_full_type_name(cls: Type["OrgsAccountResourceSpec"]) -> str:
        return f"{cls.provider_name}:{cls.type_name}"

    @classmethod
    def list_from_aws(
        cls: Type["OrgsAccountResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'ou_1_arn': {org_1_dict},
             'ou_2_arn': {org_2_dict},
             ...}

        Where the dicts represent results from list_accounts_for_parent."""
        org_resp = client.describe_organization()
        org_arn = org_resp["Organization"]["Arn"]
        # get all parents. parents include roots and ous.
        parent_ids_arns = {}
        paginator = client.get_paginator("list_roots")
        for resp in paginator.paginate():
            for root in resp["Roots"]:
                root_id, root_arn = root["Id"], root["Arn"]
                parent_ids_arns[root_id] = root_arn
                root_path = f"/{root['Name']}"
                ou_details = cls._recursively_get_ou_details_for_parent(
                    client=client, parent_id=root_id, parent_path=root_path
                )
                for ou_detail in ou_details:
                    ou_id, ou_arn = ou_detail["Id"], ou_detail["Arn"]
                    parent_ids_arns[ou_id] = ou_arn
        # now look up accounts for each ou
        orgs_accounts = {}
        accounts_paginator = client.get_paginator("list_accounts_for_parent")
        for parent_id, parent_arn in parent_ids_arns.items():
            for accounts_resp in accounts_paginator.paginate(ParentId=parent_id):
                for account in accounts_resp["Accounts"]:
                    account_id = account["Id"]
                    account_arn = f"arn:aws::::account/{account_id}"
                    account["OrganizationArn"] = org_arn
                    account["OUArn"] = parent_arn
                    orgs_accounts[account_arn] = account
        return ListFromAWSResult(resources=orgs_accounts)

    @classmethod
    def _recursively_get_ou_details_for_parent(
        cls: Type["OrgsAccountResourceSpec"], client: BaseClient, parent_id: str, parent_path: str
    ) -> List[Dict[str, Any]]:
        ous = []
        paginator = client.get_paginator("list_organizational_units_for_parent")
        for resp in paginator.paginate(ParentId=parent_id):
            for ou in resp["OrganizationalUnits"]:
                ou_id = ou["Id"]
                path = f"{parent_path}/{ou['Name']}"
                ou["Path"] = path
                ous.append(ou)
                ous += cls._recursively_get_ou_details_for_parent(
                    client=client, parent_id=ou_id, parent_path=path
                )
        return ous



"""Exceptions for access related errors."""
from altimeter.core.exceptions import AltimeterException


class AccountAuthException(AltimeterException):
    """Exception indicating auth was unable to be obtained to an account."""



"""Resource for EBSVolumes"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.instance import EC2InstanceResourceSpec
from altimeter.aws.resource.kms.key import KMSKeyResourceSpec
from altimeter.core.graph.field.dict_field import EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.resource_link_field import (
    ResourceLinkField,
    TransientResourceLinkField,
)
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class EBSVolumeResourceSpec(EC2ResourceSpec):
    """Resource for EBSVolumes"""

    type_name = "volume"
    schema = Schema(
        ScalarField("AvailabilityZone"),
        ScalarField("CreateTime"),
        ScalarField("Size"),
        ScalarField("State"),
        ScalarField("VolumeType"),
        ScalarField("Encrypted"),
        ListField(
            "Attachments",
            EmbeddedDictField(
                ScalarField("AttachTime"),
                ScalarField("State"),
                ScalarField("DeleteOnTermination"),
                ResourceLinkField("InstanceId", EC2InstanceResourceSpec),
            ),
            optional=True,
            alti_key="attachment",
        ),
        TransientResourceLinkField("KmsKeyId", KMSKeyResourceSpec, optional=True, value_is_id=True),
        TagsField(),
    )

    @classmethod
    def list_from_aws(
        cls: Type["EBSVolumeResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'volume_1_arn': {volume_1_dict},
             'volume_2_arn': {volume_2_dict},
             ...}

        Where the dicts represent results from describe_volumes."""
        volumes = {}
        paginator = client.get_paginator("describe_volumes")
        for resp in paginator.paginate():
            for volume in resp.get("Volumes", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=volume["VolumeId"]
                )
                volumes[resource_arn] = volume
        return ListFromAWSResult(resources=volumes)



"""Base class for rds resources."""

from altimeter.aws.resource.resource_spec import AWSResourceSpec


class RDSResourceSpec(AWSResourceSpec):
    """Base class for rds resources."""

    service_name = "rds"



"""Resource for EC2Images (AMIs)"""
import time
from typing import Type

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.account import AccountResourceSpec
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.core.graph.field.dict_field import EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class EC2ImageResourceSpec(EC2ResourceSpec):
    """Resource for EC2Images (AMIs)"""

    type_name = "image"
    schema = Schema(
        ScalarField("Name"),
        ScalarField("Description", optional=True),
        ScalarField("Public"),
        ListField(
            "LaunchPermissions",
            EmbeddedDictField(
                ResourceLinkField("UserId", AccountResourceSpec, optional=True, alti_key="account")
            ),
            alti_key="launch_permission",
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["EC2ImageResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'image_1_arn': {image_1_dict},
             'image_2_arn': {image_2_dict},
             ...}

        Where the dicts represent results from describe_images."""
        images = {}
        resp = client.describe_images(Owners=["self"])
        for image in resp["Images"]:
            image_id = image["ImageId"]
            time.sleep(0.25)  # seems necessary to avoid frequent RequestLimitExceeded
            try:
                perms_resp = client.describe_image_attribute(
                    Attribute="launchPermission", ImageId=image_id
                )
                launch_permissions = perms_resp["LaunchPermissions"]
                image["LaunchPermissions"] = launch_permissions
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=image_id
                )
                images[resource_arn] = image
            except ClientError as c_e:
                response_error = getattr(c_e, "response", {}).get("Error", {})
                error_code = response_error.get("Code", "")
                if error_code == "InvalidAMIID.Unavailable":
                    continue
                raise c_e
        return ListFromAWSResult(resources=images)



"""Resource representing an AWS Region"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ScanGranularity, ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class RegionResourceSpec(EC2ResourceSpec):
    """Resource representing an AWS Region"""

    type_name = "region"
    scan_granularity = ScanGranularity.ACCOUNT
    schema = Schema(ScalarField("RegionName", "name"), ScalarField("OptInStatus"))

    @classmethod
    def get_full_type_name(cls: Type["RegionResourceSpec"]) -> str:
        return f"{cls.provider_name}:{cls.type_name}"

    @classmethod
    def list_from_aws(
        cls: Type["RegionResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'region_1_arn': {region_1_dict},
             'region_2_arn': {region_2_dict},
             ...}

        Where the dicts represent results from describe_regions."""
        regions = {}
        resp = client.describe_regions(AllRegions=True)
        for region_resp in resp["Regions"]:
            region_name = region_resp["RegionName"]
            region_arn = f"arn:aws:::{account_id}:region/{region_name}"
            regions[region_arn] = region_resp
        return ListFromAWSResult(resources=regions)



"""Classes for ArtifactWriters. An ArtifactWriter writes a scan artifact dict
to something - e.g. a file, s3 key, etc."""
import abc
import io
import json
import os
from pathlib import Path
from typing import Any, Dict

import boto3

from altimeter.core.json_encoder import json_encoder
from altimeter.core.log import LogEvent, Logger


class ArtifactWriter(abc.ABC):
    """ArtifactReaders write JSON artifacts to locations - e.g. s3, filesystem, etc."""

    @abc.abstractmethod
    def write_artifact(self, name: str, data: Dict[str, Any]) -> str:
        """Read an artifact

        Args:
            name: artifact name
            data: artifact data

        Returns:
            path to written artifact
        """


class FileArtifactWriter(ArtifactWriter):
    """ArtifactWriter which writes to a file.

    Args:
         output_dir: output filesystem dir
    """

    def __init__(self, output_dir: Path):
        self.output_dir = output_dir

    def write_artifact(self, name: str, data: Dict[str, Any]) -> str:
        """Write artifact data to self.output_dir/name.json

        Args:
            name: filename
            data: artifact data

        Returns:
            Full filesystem path of artifact file
        """
        logger = Logger()
        os.makedirs(self.output_dir, exist_ok=True)
        artifact_path = os.path.join(self.output_dir, f"{name}.json")
        with logger.bind(artifact_path=artifact_path):
            logger.info(event=LogEvent.WriteToFSStart)
            with open(artifact_path, "w") as artifact_fp:
                json.dump(data, artifact_fp, default=json_encoder)
            logger.info(event=LogEvent.WriteToFSEnd)
        return artifact_path


class S3ArtifactWriter(ArtifactWriter):
    """ArtifactWriter which writes to S3.

    Args:
        bucket: s3 bucket
        key_prefix: s3 key prefix
    """

    def __init__(self, bucket: str, key_prefix: str):
        self.bucket = bucket
        self.key_prefix = key_prefix

    def write_artifact(self, name: str, data: Dict[str, Any]) -> str:
        """Write artifact data to s3://self.bucket/self.key_prefix/name.json

        Args:
            name: s3 key name
            data: artifact data

        Returns:
            S3 uri (s3://bucket/key/path) to artifact
        """

        output_key = "/".join((self.key_prefix, f"{name}.json"))
        logger = Logger()
        with logger.bind(bucket=self.bucket, key=output_key):
            logger.info(event=LogEvent.WriteToS3Start)
            s3_client = boto3.Session().client("s3")
            results_str = json.dumps(data, default=json_encoder)
            results_bytes = results_str.encode("utf-8")
            with io.BytesIO(results_bytes) as results_bytes_stream:
                s3_client.upload_fileobj(results_bytes_stream, self.bucket, output_key)
            logger.info(event=LogEvent.WriteToS3End)
        return f"s3://{self.bucket}/{output_key}"



"""ResourceSpec classes for classic elb resources."""
from altimeter.aws.resource.resource_spec import AWSResourceSpec


class ELBV1ResourceSpec(AWSResourceSpec):
    """Abstract base for ResourceSpec classes for classic elb resources."""

    service_name = "elb"

    @classmethod
    def get_client_name(cls) -> str:
        return "elb"



"""An Accessor consists of a list of MultiHopAccessors.  It provides a method `get_session`
which will iterate through the MultiHopAccessors until a session can be obtained to
a target account."""
from dataclasses import dataclass
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

import boto3
import jinja2

from altimeter.aws.auth.cache import AWSCredentialsCache
from altimeter.aws.auth.exceptions import AccountAuthException
from altimeter.aws.auth.multi_hop_accessor import MultiHopAccessor
from altimeter.core.log import LogEvent, Logger


@dataclass(frozen=True)
class Accessor:
    """An Accessor consists of a list of MultiHopAccessors.  It provides a method `get_session`
    which will iterate through the MultiHopAccessors until a session can be obtained to
    a target account.  If an Accessor has no MultiHopAccessors it simply uses the local
    session to attempt to access the account. If the session does not match the requested
    target account id, ValueError is thrown.

    Args:
        multi_hop_accessors: List of MultiHopAccessors
        credentials_cache: AWSCredentialsCache
    """

    multi_hop_accessors: List[MultiHopAccessor]
    credentials_cache: Optional[AWSCredentialsCache] = None

    def get_session(self, account_id: str, region_name: Optional[str] = None) -> boto3.Session:
        """Get a boto3 session for a given account.

        Args:
            account_id: target account id
            region_name: session region name

        Returns:
            boto3.Session object
        """
        logger = Logger()
        errors = []
        if self.multi_hop_accessors:
            for mha in self.multi_hop_accessors:  # pylint: disable=not-an-iterable
                with logger.bind(auth_accessor=str(mha)):
                    try:
                        session = mha.get_session(
                            account_id=account_id,
                            region_name=region_name,
                            credentials_cache=self.credentials_cache,
                        )
                        return session
                    except Exception as ex:
                        errors.append(ex)
                        logger.debug(event=LogEvent.AuthToAccountFailure, exception=str(ex))
            raise AccountAuthException(f"Unable to access {account_id} using {str(self)}: {errors}")
        # local run mode
        session = boto3.Session(region_name=region_name)
        sts_client = session.client("sts")
        sts_account_id = sts_client.get_caller_identity()["Account"]
        if sts_account_id != account_id:
            raise ValueError(f"BUG: sts_account_id {sts_account_id} != {account_id}")
        return session

    def __str__(self) -> str:
        return ",".join(
            [str(mha) for mha in self.multi_hop_accessors]  # pylint: disable=not-an-iterable
        )

    @classmethod
    def from_dict(cls: Type["Accessor"], data: Dict[str, Any],) -> "Accessor":
        mhas = data.get("multi_hop_accessors", [])
        credentials_cache = None
        credentials_cache_dict = data.get("credentials_cache")
        if credentials_cache_dict is not None:
            credentials_cache = AWSCredentialsCache.from_dict(credentials_cache_dict)
        return cls(
            multi_hop_accessors=[MultiHopAccessor.from_dict(mha) for mha in mhas],
            credentials_cache=credentials_cache,
        )

    @classmethod
    def from_file(cls: Type["Accessor"], filepath: Path, cache_creds: bool = True) -> "Accessor":
        """Create an Accessor from json content in a file

        Args:
            filepath: Path to json accessor definition

        Returns:
            Accessor
        """
        template_dir = str(filepath.parent.absolute())
        jinja2_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(searchpath=template_dir),
            undefined=jinja2.StrictUndefined,
        )
        template = jinja2_env.get_template(str(filepath.name))
        config = template.render(env=os.environ)
        config_dict = json.loads(config)
        if cache_creds:
            credentials_cache = AWSCredentialsCache()
            config_dict["credentials_cache"] = credentials_cache.to_dict()
        return cls.from_dict(config_dict)

    def to_dict(self) -> Dict[str, Any]:
        credentials_cache_dict = (
            None if self.credentials_cache is None else self.credentials_cache.to_dict()
        )
        return {
            "multi_hop_accessors": [mha.to_dict() for mha in self.multi_hop_accessors],
            "credentials_cache": credentials_cache_dict,
        }
